// include/strto.hh
// This file is part of libpbe; see http://svn.chezphil.org/libpbe/
// (C) 2009-2019 Philip Endecott

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef libpbe_strto_hh
#define libpbe_strto_hh

#include "moremaths.hh"

#include <stdio.h>

#include <string>
#include <cctype>
#include <algorithm>
#include <cstdint>
#include <type_traits>


namespace pbe {


struct NumberParseError: std::exception
{
  const char* what() const throw() { return "Error while parsing number"; }
};


// Decimal integer:

template <typename R, typename ITER>
R strto_unsigned(ITER p, ITER q)
{
  while (p!=q && *p==' ') {
    ++p;
  }
  if (p==q) {
    throw NumberParseError();
  }
  R r=0;
  while (p!=q && *p!=' ') {
    if (!isdigit(*p)) {
      throw NumberParseError();
    }
    r = 10*r + (*p - '0');
    ++p;
  }
  while (p!=q && *p==' ') {
    ++p;
  }
  if (p==q) {
    return r;
  }
  throw NumberParseError();
}

template <typename ITER>
uint64_t strtou64(ITER p, ITER q)
{
  return strto_unsigned<uint64_t>(p,q);
}

template <typename ITER>
unsigned long int strtoul(ITER p, ITER q)
{
  return strto_unsigned<long int>(p,q);
}

template <int N>
unsigned long int strtoul(const char (&s) [N])
{
  return strtoul(&s[0],&s[N]);
}

inline unsigned long int strtoul(std::string_view s)
{
  return strtoul(s.begin(),s.end());
}


template <typename R, typename ITER>
R strto_signed(ITER p, ITER q)
{
  while (p!=q && *p==' ') {
    ++p;
  }
  if (p==q) {
    throw NumberParseError();
  }
  int sign=1;
  if (*p=='-') {
    sign=-1;
    ++p;
  }
  typedef typename std::make_unsigned<R>::type U;
  R r = strto_unsigned<U>(p,q);
  return r * sign;
}

template <typename ITER>
long int strtol(ITER p, ITER q)
{
  while (p!=q && *p==' ') {
    ++p;
  }
  if (p==q) {
    throw NumberParseError();
  }
  int sign=1;
  if (*p=='-') {
    sign=-1;
    ++p;
  }
  unsigned long r = strtoul(p,q);
  return r * sign;
}

template <int N>
long int strtol(const char (&s) [N])
{
  return strtol(&s[0],&s[N]);
}

inline long int strtol(std::string_view s)
{
  return strtol(s.begin(),s.end());
}


// Hex integer:

template <typename ITER>
unsigned long int hex_strtoul(ITER p, ITER q)
{
  while (p<q && *p==' ') {
    ++p;
  }
  if (p==q) {
    throw NumberParseError();
  }
  unsigned long r=0;
  while (p<q && *p!=' ') {
    if (!isxdigit(*p)) {
      throw NumberParseError();
    }
    r = 16*r;
    if (isdigit(*p)) {
      r += *p - '0';
    } else {
      r += std::tolower(*p) - 'a' + 10;
    }
    ++p;
  }
  while (p<q && *p==' ') {
    ++p;
  }
  if (p==q) {
    return r;
  }
  throw NumberParseError();
}

template <int N>
unsigned long int hex_strtoul(const char (&s) [N])
{
  return hex_strtoul(&s[0],&s[N]);
}

inline unsigned long int hex_strtoul(std::string_view s)
{
  return hex_strtoul(s.begin(),s.end());
}


template <typename ITER>
long int hex_strtol(ITER p, ITER q)
{
  while (p<q && *p==' ') {
    ++p;
  }
  if (p==q) {
    throw NumberParseError();
  }
  int sign=1;
  if (*p=='-') {
    sign=-1;
    ++p;
  }
  unsigned long r = hex_strtoul(p,q);
  return r * sign;
}

template <int N>
long int hex_strtol(const char (&s) [N])
{
  return hex_strtol(&s[0],&s[N]);
}

inline long int hex_strtol(std::string_view s)
{
  return hex_strtol(s.begin(),s.end());
}


// Floating point:

template <typename T, typename ITER>
double strto_fp(ITER p, ITER q)  // FIXME surely the return type should be T
{
  // Format of the input:
  // space* [+-]? digit+ ( . digit* )? ( [eE] digit+ )? space*

  while (p != q && isspace(*p)) {
    ++p;
  }

  if (p==q) throw NumberParseError();

  bool negative = false;
  switch (*p) {
    case '-':  negative = true;  ++p; break;
    case '+':  negative = false; ++p; break;
    default:                          break;
  }

  if (p==q) throw NumberParseError();

  T n = 0;
  char c;
  do {
    c = *p;
    if (c >= '0' && c <= '9') {
      int digit = c-'0';
      n = 10*n + digit;
      ++p;
    } else if (c == '.' || c == 'e' || c == 'E' || isspace(c)) {
      ++p;
      break;
    } else {
      throw NumberParseError();
    }
  } while (p != q);

  if (c == '.') {
    T val = 0.1;
    while (p != q) {
      c = *p;
      if (c >= '0' && c <= '9') {
        int digit = c-'0';
        n = n + val*digit;
        val = val * 0.1;
        ++p;
      } else if (c == 'e' || c == 'E' || isspace(c)) {
        ++p;
        break;
      } else {
        throw NumberParseError();
      }
    }
  }

  if (c == 'e' || c == 'E') {
    if (p==q) throw NumberParseError();
    bool neg_expo = false;
    switch (*p) {
      case '-':  neg_expo = true;  ++p; break;
      case '+':  neg_expo = false; ++p; break;
      default:                          break;
    }
    if (p==q) throw NumberParseError();
    int expo = 0;
    do {
      c = *p;
      if (c >= '0' && c <= '9') {
        int digit = c-'0';
        expo = 10*expo + digit;
        ++p;
      } else if (isspace(c)) {
        ++p;
        break;
      } else {
        throw NumberParseError();
      }
    } while (p != q);
    if (neg_expo) expo = -expo;
    n = n * exp10(static_cast<T>(expo));
  }

  while (p != q) {
    if (!isspace(*p)) {
      throw NumberParseError();
    }
    ++p;
  }

  if (negative) n = -n;

  return n;
}

template <typename ITER>
double strtod(ITER p, ITER q)
{
  return strto_fp<double>(p,q);
}

template <int N>
double strtod(const char (&s) [N])
{
  return strtod(&s[0],&s[N]);
}

inline double strtod(std::string_view s)
{
  return strtod(s.begin(),s.end());
}


template <typename ITER>
double strtof(ITER p, ITER q)
{
  return strto_fp<float>(p,q);
}

template <int N>
float strtof(const char (&s) [N])
{
  return strtof(&s[0],&s[N]);
}

inline float strtof(std::string_view s)
{
  return strtof(s.begin(),s.end());
}


// Template version:


// These helpers are distinguished by a final parameter that indicates 
// the required type; they forward to the explicitly-named implementations 
// above:

template <typename ITER> uint64_t strto(ITER p, ITER q, uint64_t) { return strto_unsigned<uint64_t>(p,q); }
template <typename ITER>  int64_t strto(ITER p, ITER q,  int64_t) { return strto_signed   <int64_t>(p,q); }
template <typename ITER> uint32_t strto(ITER p, ITER q, uint32_t) { return strto_unsigned<uint32_t>(p,q); }
template <typename ITER>  int32_t strto(ITER p, ITER q,  int32_t) { return strto_signed   <int32_t>(p,q); }
template <typename ITER> uint16_t strto(ITER p, ITER q, uint16_t) { return strto_unsigned<uint16_t>(p,q); }
template <typename ITER>  int16_t strto(ITER p, ITER q,  int16_t) { return strto_signed   <int16_t>(p,q); }
template <typename ITER>   double strto(ITER p, ITER q,   double) { return strtod (p,q); }
template <typename ITER>    float strto(ITER p, ITER q,    float) { return strtof (p,q); }


// This dispatches to one of the above based on the supplied template 
// parameter type:

template <typename T, typename ITER>
T strto(ITER p, ITER q)
{
  return strto(p,q,T());
}

// These forward to the above:

template <typename T, int N>
T strto(const char (&s) [N])
{
  return strto<T>(&s[0],&s[N]);
}

template <typename T>
T strto(std::string_view s)
{
  return strto<T>(s.begin(),s.end());
}

template <typename T>
T strto(char c)
{
  return strto<T>(&c,(&c)+1);
}


// As above, for hex:

//template <typename ITER>           uint64_t hex_strto(ITER p, ITER q,           uint64_t) { return hex_strtou64(p,q); }
//template <typename ITER>            int64_t hex_strto(ITER p, ITER q,            int64_t) { return hex_strto64 (p,q); }
template <typename ITER> unsigned long  int hex_strto(ITER p, ITER q, unsigned long  int) { return hex_strtoul(p,q); }
template <typename ITER>          long  int hex_strto(ITER p, ITER q,          long  int) { return hex_strtol (p,q); }
template <typename ITER> unsigned       int hex_strto(ITER p, ITER q, unsigned       int) { return hex_strtoul(p,q); }
template <typename ITER>                int hex_strto(ITER p, ITER q,                int) { return hex_strtol (p,q); }
template <typename ITER> unsigned short int hex_strto(ITER p, ITER q, unsigned short int) { return hex_strtoul(p,q); }
template <typename ITER>          short int hex_strto(ITER p, ITER q,          short int) { return hex_strtol (p,q); }

template <typename T, typename ITER>
T hex_strto(ITER p, ITER q)
{
  return hex_strto(p,q,T());
}

template <typename T, int N>
T hex_strto(const char (&s) [N])
{
  return hex_strto<T>(&s[0],&s[N]);
}

template <typename T>
T hex_strto(std::string_view s)
{
  return hex_strto<T>(s.begin(),s.end());
}


};


#endif


